import platform
import resource
from pathlib import Path
from typing import Any

import yaml


async def enumerate_async(iterable):
    count = 0

    async for item in iterable:
        yield count, item
        count += 1


def get_max_mem_mb():
    max_mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    max_mem /= 1024
    if platform.system() == "Darwin":
        max_mem /= 1024
    return int(max_mem)


def pretty_print_yaml_to_file(path: Path, data: Any):
    with open(path, "w") as fp:
        yaml.safe_dump(data, fp, indent=2, sort_keys=True)



from functools import reduce
from typing import Any, AsyncGenerator, Iterable, List

from ..model import AggregatedIntrospectionMixin, IntrospectiveIngestionComponent
from .step import Step


async def empty_async_generator():
    for item in []:
        yield item


class Pipeline(AggregatedIntrospectionMixin, IntrospectiveIngestionComponent):
    """A pipeline is a series of steps that are executed in order."""

    __slots__ = ("steps",)

    def __init__(self, steps: List[Step]) -> None:
        self.steps = steps

    def run(self) -> AsyncGenerator[Any, Any]:
        record_stream_over_all_steps = reduce(
            lambda stream, step: step.handle_async_record_stream(stream),
            self.steps,
            empty_async_generator(),
        )
        return record_stream_over_all_steps

    def all_subordinate_components(self) -> Iterable[IntrospectiveIngestionComponent]:
        return (s for s in self.steps if isinstance(s, IntrospectiveIngestionComponent))



from pathlib import Path

from cleo.helpers import argument

from ..operations import (
    CommitProjectToDisk,
    GeneratePipelineScaffold,
    GenerateProject,
    GeneratePythonScaffold,
)
from .nodestream_command import NodestreamCommand
from .shared_options import DATABASE_NAME_OPTION


class New(NodestreamCommand):
    name = "new"
    description = "Generate a New Nodestream"
    arguments = [
        argument("project", "the name of the project to create"),
    ]
    options = [DATABASE_NAME_OPTION]

    async def handle_async(self):
        project_root = Path(self.argument("project"))
        db = self.option("database")
        python_files = await self.run_operation(GeneratePythonScaffold(project_root))
        pipelines = await self.run_operation(GeneratePipelineScaffold(project_root, db))
        gen_project = GenerateProject(project_root, pipelines, python_files, db)
        project = await self.run_operation(gen_project)
        commit_to_disk = CommitProjectToDisk(project, project_root / "nodestream.yaml")
        await self.run_operation(commit_to_disk)



from dataclasses import dataclass
from typing import Any, Dict, List

COMMIT_QUERY = """
CALL apoc.periodic.iterate(
    "UNWIND $batched_parameter_sets as params RETURN params",
    $batched_query,
    {batchsize: 1000, paralell: false, retries: 3, params: {batched_parameter_sets: $batched_parameter_sets}}
)
YIELD batches, committedOperations, failedOperations, errorMessages
RETURN batches, committedOperations, failedOperations, errorMessages
"""


@dataclass(slots=True)
class Query:
    query_statement: str
    parameters: Dict[str, Any]

    @classmethod
    def from_statement(cls, statement: str):
        return cls(query_statement=statement, parameters={})


@dataclass(slots=True)
class QueryBatch:
    query_statement: str
    batched_parameter_sets: List[Dict[str, Any]]

    def as_query(self) -> Query:
        return Query(
            COMMIT_QUERY,
            {
                "batched_parameter_sets": self.batched_parameter_sets,
                "batched_query": self.query_statement,
            },
        )



from .audit import Audit
from .audit_printer import AuditPrinter
from .audit_referencial_integrity import AuditReferencialIntegrity
from .audit_ttls import TTLAudit

__all__ = (
    "Audit",
    "AuditPrinter",
    "TTLAudit",
    "AuditReferencialIntegrity",
)



from contextlib import contextmanager
from contextvars import ContextVar

UNKNOWN_PIPELINE_NAME = "unknown"

pipeline_name: ContextVar[str] = ContextVar("pipeline_name")


def get_pipeline_name() -> str:
    return pipeline_name.get(UNKNOWN_PIPELINE_NAME)


@contextmanager
def set_pipeline_name(name: str):
    token = pipeline_name.set(name)
    try:
        yield
    finally:
        pipeline_name.reset(token)



from typing import Any, Dict

from ..exceptions import UnhandledBranchError
from ..model import AggregatedIntrospectionMixin, InterpreterContext
from ..value_providers import StaticValueOrValueProvider, ValueProvider
from .interpretation import Interpretation


class SwitchInterpretation(
    AggregatedIntrospectionMixin, Interpretation, alias="switch"
):
    __slots__ = (
        "switch_on",
        "interpretations",
        "default",
        "normalization",
        "fail_on_unhandled",
    )

    def __init__(
        self,
        switch_on: StaticValueOrValueProvider,
        cases: Dict[str, dict],
        default: Dict[str, Any] = None,
        normalization: Dict[str, Any] = None,
        fail_on_unhandled: bool = True,
    ):
        self.switch_on = ValueProvider.guarantee_value_provider(switch_on)
        self.interpretations = {
            field_value: Interpretation.from_file_arguments(**interpretation)
            for field_value, interpretation in cases.items()
        }
        self.default = (
            Interpretation.from_file_arguments(**default) if default else None
        )
        self.normalization = normalization or {}
        self.fail_on_unhandled = fail_on_unhandled

    def all_subordinate_components(self):
        yield from self.interpretations.values()
        if self.default:
            yield self.default

    def interpret(self, context: InterpreterContext):
        value_to_look_for = self.switch_on.normalize_single_value(
            context, **self.normalization
        )
        if value_to_look_for not in self.interpretations:
            if self.default:
                self.default.interpret(context)
                return
            if self.fail_on_unhandled:
                raise UnhandledBranchError(value_to_look_for)
            else:
                return
        return self.interpretations[value_to_look_for].interpret(context)



from .desired_ingest import DesiredIngestion, MatchStrategy, RelationshipWithNodes
from .graph_objects import (
    Node,
    NodeIdentityShape,
    PropertySet,
    Relationship,
    RelationshipIdentityShape,
)
from .indexes import FieldIndex, KeyIndex
from .ingest_strategy import IngestionStrategy
from .ingestion_hooks import IngestionHook, IngestionHookRunRequest
from .interpreter_context import InterpreterContext, JsonLikeDocument
from .schema import (
    AggregatedIntrospectionMixin,
    Cardinality,
    GraphObjectShape,
    GraphObjectType,
    GraphSchema,
    IntrospectiveIngestionComponent,
    KnownTypeMarker,
    PresentRelationship,
    PropertyMetadata,
    PropertyMetadataSet,
    PropertyType,
    TypeMarker,
    UnknownTypeMarker,
)
from .ttl import TimeToLiveConfiguration

__all__ = (
    "DesiredIngestion",
    "RelationshipWithNodes",
    "MatchStrategy",
    "PropertySet",
    "Node",
    "Relationship",
    "KeyIndex",
    "FieldIndex",
    "IngestionStrategy",
    "IngestionHook",
    "IngestionHookRunRequest",
    "TimeToLiveConfiguration",
    "InterpreterContext",
    "RecordDecomposer",
    "JsonLikeDocument",
    "GraphObjectShape",
    "GraphObjectType",
    "TypeMarker",
    "KnownTypeMarker",
    "UnknownTypeMarker",
    "PropertyMetadataSet",
    "PropertyMetadata",
    "PropertyType",
    "IntrospectiveIngestionComponent",
    "AggregatedIntrospectionMixin",
    "Cardinality",
    "PresentRelationship",
    "RelationshipIdentityShape",
    "NodeIdentityShape",
    "GraphSchema",
)



from typing import Optional

from ..model.ingest_strategy import INGESTION_STRATEGY_REGISTRY, IngestionStrategy
from ..pipeline import Flush, Writer
from .debounced_ingest_strategy import DebouncedIngestStrategy
from .query_executor import QUERY_EXECUTOR_SUBCLASS_REGISTRY


class GraphDatabaseWriter(Writer):
    @classmethod
    def __declarative_init__(
        cls,
        batch_size: int,
        database: str,
        ingest_strategy_name: Optional[str] = None,
        **database_args
    ):
        ingest_strategy_name = (
            ingest_strategy_name
            or INGESTION_STRATEGY_REGISTRY.name_for(DebouncedIngestStrategy)
        )
        ingest_strategy = INGESTION_STRATEGY_REGISTRY.get(ingest_strategy_name)
        executor_class = QUERY_EXECUTOR_SUBCLASS_REGISTRY.get(database)
        executor = executor_class.from_file_arguments(**database_args)
        return cls(
            batch_size=batch_size,
            ingest_strategy=ingest_strategy(executor),
        )

    def __init__(self, batch_size: int, ingest_strategy: IngestionStrategy) -> None:
        self.batch_size = batch_size
        self.ingest_strategy = ingest_strategy
        self.pending_records = 0

    async def flush(self):
        await self.ingest_strategy.flush()
        self.pending_records = 0

    async def write_record(self, ingestible):
        if ingestible is Flush:
            await self.flush()
            return

        await ingestible.ingest(self.ingest_strategy)
        self.pending_records += 1
        if self.pending_records >= self.batch_size:
            await self.flush()

    async def finish(self):
        await self.flush()



from abc import abstractmethod
from typing import Any, AsyncGenerator, Generator

from .flush import Flush
from .step import Step


class Transformer(Step):
    """A `Transformer` takes a given record and mutates into a new record.

    `Transformer` steps generally make up the middle of an ETL pipeline and are responsible
    for reshaping an object so its more ingestible by the downstream sink.
    """

    async def handle_async_record_stream(
        self, record_stream: AsyncGenerator[Any, Any]
    ) -> AsyncGenerator[Any, Any]:
        async for record in record_stream:
            if record is Flush:
                yield record
            else:
                val_or_gen = await self.transform_record(record)
                if isinstance(val_or_gen, Generator):
                    for result in val_or_gen:
                        yield result
                else:
                    yield val_or_gen

    @abstractmethod
    async def transform_record(self, record: Any) -> Any:
        raise NotImplementedError



from ...pipeline import PipelineInitializationArguments
from ...project import PipelineProgressReporter, Project, RunRequest
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation


class RunPipeline(Operation):
    def __init__(self, project: Project) -> None:
        self.project = project

    async def perform(self, command: NodestreamCommand):
        await self.project.run(self.make_run_request(command))

    def make_run_request(self, command: NodestreamCommand) -> RunRequest:
        return RunRequest(
            pipeline_name=command.argument("pipeline"),
            initialization_arguments=PipelineInitializationArguments(
                annotations=command.option("annotations"),
            ),
            progress_reporter=self.create_progress_reporter(command),
        )

    def get_progress_indicator(self, command: NodestreamCommand) -> "ProgressIndicator":
        if command.has_json_logging_set:
            return ProgressIndicator(command)

        return SpinnerProgressIndicator(command)

    def create_progress_reporter(
        self, command: NodestreamCommand
    ) -> PipelineProgressReporter:
        indicator = self.get_progress_indicator(command)
        return PipelineProgressReporter(
            reporting_frequency=int(command.option("reporting-frequency")),
            callback=indicator.progress_callback,
            on_start_callback=indicator.on_start,
            on_finish_callback=indicator.on_finish,
        )


class ProgressIndicator:
    def __init__(self, command: NodestreamCommand) -> None:
        self.command = command

    def on_start(self):
        pass

    def progress_callback(self, _, __):
        pass

    def on_finish(self):
        pass

    @property
    def pipeline_name(self) -> str:
        return self.command.argument("pipeline")


class SpinnerProgressIndicator(ProgressIndicator):
    def on_start(self):
        self.progress = self.command.progress_indicator()
        self.progress.start(f"Running pipeline: '{self.pipeline_name}'")

    def progress_callback(self, index, _):
        self.progress.set_message(
            f"Currently processing record at index: <info>{index}</info>"
        )

    def on_finish(self):
        self.progress.finish(f"Finished running pipeline: '{self.pipeline_name}'")



from typing import Dict, Iterable, List

from ..exceptions import MissingExpectedPipelineError
from ..model import AggregatedIntrospectionMixin, IntrospectiveIngestionComponent
from .pipeline_definition import PipelineDefinition
from .run_request import RunRequest


class PipelineScope(AggregatedIntrospectionMixin, IntrospectiveIngestionComponent):
    """A `PipelineScope` represents a collection of pipelines subordinate to a project."""

    def __init__(self, name: str, pipelines: List[PipelineDefinition]) -> None:
        self.name = name
        self.pipelines_by_name: Dict[str, PipelineDefinition] = {}
        for pipeline in pipelines:
            self.add_pipeline_definition(pipeline)

    @classmethod
    def from_file_data(cls, scope_name, file_data):
        pipelines_data = file_data.pop("pipelines", [])
        annotations = file_data.pop("annotations", {})
        pipelines = [
            PipelineDefinition.from_file_data(pipeline_data, annotations)
            for pipeline_data in pipelines_data
        ]
        return cls(scope_name, pipelines)

    async def run_request(self, run_request: "RunRequest"):
        if (name := run_request.pipeline_name) not in self:
            return

        await run_request.execute_with_definition(self[name])

    def __getitem__(self, pipeline_name):
        return self.pipelines_by_name[pipeline_name]

    def __contains__(self, pipeline_name):
        return pipeline_name in self.pipelines_by_name

    def add_pipeline_definition(self, definition: PipelineDefinition):
        self.pipelines_by_name[definition.name] = definition

    def delete_pipeline(
        self,
        pipeline_name: str,
        remove_pipeline_file: bool = True,
        missing_ok: bool = True,
    ) -> bool:
        definition = self.pipelines_by_name.get(pipeline_name)
        if definition is None:
            if not missing_ok:
                raise MissingExpectedPipelineError(
                    "Attempted to delete pipeline that did not exist"
                )
            else:
                return False

        del self.pipelines_by_name[pipeline_name]
        if remove_pipeline_file:
            definition.remove_file(missing_ok=missing_ok)

        return True

    def as_dict(self):
        return {
            "pipelines": [
                ppl.as_file_definition() for ppl in self.pipelines_by_name.values()
            ]
        }

    def all_subordinate_components(self) -> Iterable[IntrospectiveIngestionComponent]:
        return self.pipelines_by_name.values()



from collections import defaultdict
from typing import Dict, Iterable, List, Tuple

from ..model import MatchStrategy, Node, RelationshipWithNodes
from .query_executor import OperationOnNodeIdentity, OperationOnRelationshipIdentity


class OperationDebouncer:
    def __init__(self):
        self.nodes_by_shape_operation: Dict[
            OperationOnNodeIdentity, List[Node]
        ] = defaultdict(list)
        self.relationships_by_operation: Dict[
            OperationOnRelationshipIdentity, List[RelationshipWithNodes]
        ] = defaultdict(list)

    def bucketize_node_operation(
        self, node: Node, match_strategy: MatchStrategy
    ) -> List[Node]:
        return self.nodes_by_shape_operation[
            OperationOnNodeIdentity(node.identity_shape, match_strategy)
        ]

    def bucketize_relationship_operation(
        self, relationship: RelationshipWithNodes
    ) -> List[RelationshipWithNodes]:
        return self.relationships_by_operation[
            OperationOnRelationshipIdentity(
                from_node=OperationOnNodeIdentity(
                    node_identity=relationship.from_node.identity_shape,
                    match_strategy=relationship.from_side_match_strategy.prevent_creation(),
                ),
                to_node=OperationOnNodeIdentity(
                    node_identity=relationship.to_node.identity_shape,
                    match_strategy=relationship.to_side_match_strategy.prevent_creation(),
                ),
                relationship_identity=relationship.relationship.identity_shape,
            )
        ]

    def debounce_node_operation(
        self, node: Node, match_strategy: MatchStrategy = MatchStrategy.EAGER
    ):
        bucket = self.bucketize_node_operation(node, match_strategy)
        for existing_node in bucket:
            if existing_node.has_same_key(node):
                existing_node.update(node)
                break
        else:
            bucket.append(node)

    def debounce_relationship(self, relationship: RelationshipWithNodes):
        bucket = self.bucketize_relationship_operation(relationship)
        for existing_rel in bucket:
            if existing_rel.has_same_keys(relationship):
                existing_rel.update(relationship)
                break
        else:
            bucket.append(relationship)

    def drain_node_groups(
        self,
    ) -> Iterable[Tuple[OperationOnNodeIdentity, List[Node]]]:
        yield from self.nodes_by_shape_operation.items()
        self.nodes_by_shape_operation.clear()

    def drain_relationship_groups(
        self,
    ) -> Iterable[Tuple[OperationOnRelationshipIdentity, List[RelationshipWithNodes]]]:
        yield from self.relationships_by_operation.items()
        self.relationships_by_operation.clear()



from typing import Type

from ...audits import Audit
from ..operations import InitializeProject, RunAudit
from .nodestream_command import NodestreamCommand
from .shared_options import PROJECT_FILE_OPTION


class AuditCommand(NodestreamCommand):
    audit: Type[Audit] = Audit
    options = [PROJECT_FILE_OPTION]

    async def handle_async(self):
        project = await self.run_operation(InitializeProject())
        return await self.run_operation(RunAudit(project, self.audit))



import json
from abc import ABC, abstractmethod
from contextlib import contextmanager
from csv import DictReader
from glob import glob
from io import StringIO
from pathlib import Path
from typing import Any, AsyncGenerator, Iterable, Union

from ..model import JsonLikeDocument
from ..pipeline import Extractor
from ..subclass_registry import SubclassRegistry

SUPPORTED_FILE_FORMAT_REGISTRY = SubclassRegistry()


@SUPPORTED_FILE_FORMAT_REGISTRY.connect_baseclass
class SupportedFileFormat(ABC):
    def __init__(self, file: Union[Path, StringIO]) -> None:
        self.file = file

    @contextmanager
    def read_handle(self) -> StringIO:
        if isinstance(self.file, Path):
            with open(self.file, "r+") as fp:
                yield fp
        else:
            yield self.file

    def read_file(self) -> Iterable[JsonLikeDocument]:
        with self.read_handle() as fp:
            return self.read_file_from_handle(fp)

    @classmethod
    def open(cls, file: Path) -> "SupportedFileFormat":
        return cls.from_file_pointer_and_format(file, file.suffix)

    @classmethod
    def from_file_pointer_and_format(
        cls, fp: StringIO, file_format: str
    ) -> "SupportedFileFormat":
        file_format = SUPPORTED_FILE_FORMAT_REGISTRY.get(file_format)
        return file_format(fp)

    @abstractmethod
    def read_file_from_handle(self, fp: StringIO) -> Iterable[JsonLikeDocument]:
        ...


class JsonFileFormat(SupportedFileFormat, alias=".json"):
    def read_file_from_handle(self, fp: StringIO) -> Iterable[JsonLikeDocument]:
        return [json.load(fp)]


class TextFileFormat(SupportedFileFormat, alias=".txt"):
    def read_file_from_handle(self, fp: StringIO) -> Iterable[JsonLikeDocument]:
        return [{"line": line} for line in fp.readlines()]


class CommaSeperatedValuesFileFormat(SupportedFileFormat, alias=".csv"):
    def read_file_from_handle(self, fp: StringIO) -> Iterable[JsonLikeDocument]:
        return tuple(DictReader(fp))


class FileExtractor(Extractor):
    @classmethod
    def __declarative_init__(cls, globs: Iterable[str]):
        all_matching_paths = (
            Path(file)
            for glob_string in globs
            for file in glob(glob_string, recursive=True)
        )
        final_paths = {p for p in all_matching_paths if p.is_file()}
        return cls(final_paths)

    def __init__(self, paths: Iterable[Path]) -> None:
        self.paths = paths

    async def extract_records(self) -> AsyncGenerator[Any, Any]:
        for path in self.paths:
            for record in SupportedFileFormat.open(path).read_file():
                yield record



from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

from yaml import SafeLoader, load

from ..argument_resolvers import ARGUMENT_RESOLVER_REGISTRY
from ..exceptions import InvalidPipelineDefinitionError
from ..value_providers import VALUE_PROVIDER_REGISTRY
from .class_loader import ClassLoader
from .pipeline import Pipeline


class PipelineFileSafeLoader(SafeLoader):
    """A YAML loader that can load pipeline files.""" ""

    was_configured = False

    @classmethod
    def configure(cls):
        if cls.was_configured:
            return

        for value_provider in VALUE_PROVIDER_REGISTRY.all_subclasses:
            value_provider.install_yaml_tag(cls)
        for argument_resolver in ARGUMENT_RESOLVER_REGISTRY.all_subclasses:
            argument_resolver.install_yaml_tag(cls)

        cls.was_configured = True

    @classmethod
    def load_file_by_path(cls, file_path: str):
        PipelineFileSafeLoader.configure()
        with open(file_path) as fp:
            return load(fp, cls)


@dataclass(slots=True)
class PipelineInitializationArguments:
    """Arguments used to initialize a pipeline from a file."""

    annotations: Optional[List[str]] = None

    @classmethod
    def for_introspection(cls):
        return cls(annotations=["introspection"])

    def initialize_from_file_data(self, file_data: List[dict]):
        return Pipeline(self.load_steps(ClassLoader(), file_data))

    def load_steps(self, class_loader, file_data):
        return [
            class_loader.load_class(**step_data)
            for step_data in file_data
            if self.should_load_step(step_data)
        ]

    def should_load_step(self, step):
        return self.step_is_tagged_properly(step)

    def step_is_tagged_properly(self, step):
        if "annotations" in step and self.annotations is not None:
            if not set(step.pop("annotations")).intersection(self.annotations):
                return False

        return True


class PipelineFileLoader:
    """Loads a pipeline from a YAML file."""

    def __init__(self, file_path: Path):
        self.file_path = file_path

    def load_pipeline(
        self, init_args: Optional[PipelineInitializationArguments] = None
    ) -> Pipeline:
        init_args = init_args or PipelineInitializationArguments()
        return self.load_pipeline_from_file_data(
            self.load_pipeline_file_data(), init_args
        )

    def load_pipeline_from_file_data(
        self, file_data, init_args: PipelineInitializationArguments
    ):
        if not isinstance(file_data, list):
            raise InvalidPipelineDefinitionError(
                "File should be a list of step class to load"
            )

        return init_args.initialize_from_file_data(file_data)

    def load_pipeline_file_data(self):
        return PipelineFileSafeLoader.load_file_by_path(self.file_path)



from typing import Any

from ...logging import configure_logging_with_json_defaults
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation


class InitializeLogger(Operation):
    async def perform(self, command: NodestreamCommand) -> Any:
        if command.has_json_logging_set:
            configure_logging_with_json_defaults()



from dataclasses import dataclass, field
from logging import Logger, getLogger
from typing import Any, Callable

from ..pipeline import Pipeline
from ..utilities import enumerate_async, get_max_mem_mb


def no_op(*_, **__):
    pass


@dataclass
class PipelineProgressReporter:
    """A `PipelineProgressReporter` is a utility that can be used to report on the progress of a pipeline."""

    reporting_frequency: int = 1000
    logger: Logger = field(default_factory=getLogger)
    callback: Callable[[int, Any], None] = field(default=no_op)
    on_start_callback: Callable[[], None] = field(default=no_op)
    on_finish_callback: Callable[[], None] = field(default=no_op)

    async def execute_with_reporting(self, pipeline: Pipeline):
        self.on_start_callback()
        async for index, record in enumerate_async(pipeline.run()):
            if index % self.reporting_frequency == 0:
                self.report(index, record)
        self.on_finish_callback()

    def report(self, index, record):
        self.logger.info(
            "Records Processed", extra={"index": index, "max_memory": get_max_mem_mb()}
        )
        self.callback(index, record)



import os
import sys

from cleo.application import Application

from .commands.audit_ttls import AuditTTLs
from .commands.audit_refs import AuditRefs
from .commands.new import New
from .commands.print_schema import PrintSchema
from .commands.remove import Remove
from .commands.run import Run
from .commands.scaffold import Scaffold
from .commands.show import Show

APPLICATION = Application(name="nodestream")
APPLICATION.add(Run())
APPLICATION.add(New())
APPLICATION.add(Show())
APPLICATION.add(Scaffold())
APPLICATION.add(Remove())
APPLICATION.add(PrintSchema())
APPLICATION.add(AuditTTLs())
APPLICATION.add(AuditRefs())


def run():
    # For installed users, the current working directory is not guaranteed to be
    # in the python path. So, we will guarantee it is by adding it to path before we
    # run the application.
    sys.path.append(os.getcwd())
    APPLICATION.run()


if __name__ == "__main__":
    run()



from .pipeline_definition import PipelineDefinition
from .pipeline_progress_reporter import PipelineProgressReporter
from .pipeline_scope import PipelineScope
from .project import Project
from .run_request import RunRequest

__all__ = (
    "PipelineDefinition",
    "PipelineProgressReporter",
    "PipelineScope",
    "Project",
    "RunRequest",
)



from io import StringIO
from pathlib import Path
from typing import Any, AsyncGenerator, Optional

from ....pipeline import Extractor
from ...files import SupportedFileFormat
from .credential_utils import AwsClientFactory


class S3Extractor(Extractor):
    @classmethod
    def __declarative_init__(
        cls,
        bucket: str,
        prefix: Optional[str] = None,
        object_format: Optional[str] = None,
        **aws_client_args,
    ):
        return cls(
            bucket=bucket,
            object_format=object_format,
            prefix=prefix,
            s3_client=AwsClientFactory(**aws_client_args).make_client("s3"),
        )

    def __init__(
        self,
        bucket: str,
        s3_client,
        object_format: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> None:
        self.object_format = object_format
        self.prefix = prefix or ""
        self.bucket = bucket
        self.s3_client = s3_client

    def get_object_as_io(self, key: str) -> StringIO:
        return self.s3_client.get_object(Bucket=self.bucket, Key=key)["Body"]

    def infer_object_format(self, key: str) -> str:
        object_format = self.object_format or Path(key).suffix
        if not object_format:
            raise ValueError(
                f"No object format provided and key has no extension: '{key}'"
            )
        return object_format

    def get_object_as_file(self, key: str) -> SupportedFileFormat:
        io = self.get_object_as_io(key)
        object_format = self.infer_object_format(key)
        return SupportedFileFormat.from_file_pointer_and_format(io, object_format)

    def find_keys_in_bucket(self) -> list[str]:
        paginator = self.s3_client.get_paginator("list_objects_v2")
        page_iterator = paginator.paginate(Bucket=self.bucket, Prefix=self.prefix)
        for page in page_iterator:
            for obj in page["Contents"]:
                yield obj["Key"]

    async def extract_records(self) -> AsyncGenerator[Any, Any]:
        for key in self.find_keys_in_bucket():
            for record in self.get_object_as_file(key).read_file():
                yield record



from pathlib import Path


class InvalidClassPathError(ValueError):
    """Raised when a class path is invalid."""

    pass


class InvalidPipelineDefinitionError(ValueError):
    """Raised when a pipeline definition is invalid."""

    pass


class InvalidPipelineFileError(ValueError):
    """Raised when a pipeline file is invalid."""

    pass


class PipelineComponentInitializationError(ValueError):
    """Raised when a component fails to initialize."""

    def __init__(self, initializer, init_arguments, *args: object) -> None:
        super().__init__(
            "Failed to Initialize Component in Declarative Pipeline. Likely the arguments are incorrect.",
            *args,
        )
        self.initializer = initializer
        self.init_arguments = init_arguments


class AlreadyInRegistryError(ValueError):
    """Raised when a subclass with the same name is already in the subclass registry."""

    def __init__(self, name, *args: object) -> None:
        super().__init__(f"{name} is already registered", *args)


class MissingFromRegistryError(ValueError):
    """Raised when a subclass is not in the subclass registry."""

    def __init__(self, name, *args: object) -> None:
        super().__init__(f"{name} is not in the subclass registry", *args)


class InvalidKeyLengthError(ValueError):
    """Raised when a related nodes have differing lengths of key parts returned from a value provider.."""

    def __init__(self, district_lengths, *args: object) -> None:
        lengths = f"({','.join((str(length) for length in district_lengths))})"
        error = f"Node Relationships do not have a consistent key length. Lengths are: ({lengths}) "
        super().__init__(error, *args)


class InvalidFlagError(ValueError):
    """Raised when a normalization flag is not valid."""

    def __init__(self, flag_name, *args: object) -> None:
        super().__init__(
            f"Normalization flag with name '{flag_name}' is not valid.`", *args
        )


class UnhandledBranchError(ValueError):
    """Raised when a branch is not handled in a switch case."""

    def __init__(self, missing_branch_value, *args: object) -> None:
        super().__init__(
            f"'{missing_branch_value}' was not matched in switch case", *args
        )


class MissingProjectFileError(ValueError):
    """Raised when a project file is missing."""

    def __init__(self, file: Path, *args: object) -> None:
        super().__init__(f"'{file}' does not exist", *args)


class MissingExpectedPipelineError(ValueError):
    pass



from typing import Any, Iterable, Type

import jq
from yaml import SafeLoader

from ..model import InterpreterContext
from .value_provider import ValueProvider


class JqValueProvider(ValueProvider):
    """A `ValueProvider` that uses Jq to extract values from a document."""

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        loader.add_constructor(
            "!jq", lambda loader, node: cls(jq.compile(loader.construct_scalar(node)))
        )

    def __init__(self, jq_program) -> None:
        self.jq_program = jq_program

    def search(self, context: InterpreterContext):
        raw_search = self.jq_program.input(context.document).all()
        for hit in raw_search:
            if hit is None:
                return
            if isinstance(hit, list):
                yield from hit
            else:
                yield hit

    def single_value(self, context: InterpreterContext) -> Any:
        return next(self.search(context), None)

    def many_values(self, context: InterpreterContext) -> Iterable[Any]:
        return self.search(context)



from ...audits import AuditReferencialIntegrity
from .audit_command import AuditCommand


class AuditRefs(AuditCommand):
    name = "audit refs"
    description = "Audit the correctness of references between nodes of the project"
    audit = AuditReferencialIntegrity



from enum import Enum


class MatchStrategy(str, Enum):
    EAGER = "EAGER"
    MATCH_ONLY = "MATCH_ONLY"
    FUZZY = "FUZZY"

    def prevent_creation(self) -> "MatchStrategy":
        if self == MatchStrategy.EAGER:
            return MatchStrategy.MATCH_ONLY
        return self
